#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>

#include <stdio.h> // perror
#include <unistd.h> // close
#include <string.h> // memset

#include "socket.hh"

static std::string PeerIp;

int SockAccept(int fd, bool blocking)
{
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);
    int sfd;

    if ((sfd = accept(fd, (struct sockaddr *)&addr, &addrlen)) == -1) {
        if (errno == EAGAIN || errno == EWOULDBLOCK) {
            return -1;
        } else {
            perror("accept()");
        }
        return -1;
    }
    
    if(!blocking)
    {
        int flags = -1;
        if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
            fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0) {
            perror("setting O_NONBLOCK");
            close(sfd);
            return -1;
        }
    }
    
    PeerIp = inet_ntoa(addr.sin_addr);
    
    return sfd;
}

const std::string& GetLastPeerIp()
{
    return PeerIp;
}

int SockAcceptNonblock(int fd)
{
    return SockAccept(fd, false);
}

int SockAcceptDoblock(int fd)
{
    return SockAccept(fd, true);
}

static int new_socket(bool blocking)
{
    int sfd;

    if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket()");
        return -1;
    }

    if(!blocking)
    {
        int flags;
        if ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
            fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0) {
            perror("setting O_NONBLOCK");
            close(sfd);
            return -1;
        }
    }
    return sfd;
}

int MakeServerSocket(int port, bool blocking)
{
    int sfd;
    struct linger ling = {0, 0};
    struct sockaddr_in addr;
    int flags =1;

    if ((sfd = new_socket(blocking)) == -1) {
        return -1;
    }

    setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
    setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
    setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));
#if !defined(TCP_NOPUSH)
    setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(flags));
#endif

    /* 
     * the memset call clears nonstandard fields in some impementations
     * that otherwise mess things up.
     */
    memset(&addr, 0, sizeof(addr));

    struct in_addr interface;
    interface.s_addr = htonl(INADDR_ANY);
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr = interface;
    if (bind(sfd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
        perror("bind()");
        close(sfd);
        return -1;
    }
    if (listen(sfd, 1024) == -1) {
        perror("listen()");
        close(sfd);
        return -1;
    }
    return sfd;
}
